#include "stdafx.h"
#include "clf_util.h"
#include "clf_random.h"
#include "clf_primitives.h"
#include "clf_geometry.h"
#include "clf_camera.h"
#include "clf_resources.h"
#include "clf_light.h"
#include "clf_font.h"
#include "clf_shader.h"

extern HWND  hWnd;						// Window Handle
extern bool keyDown[256];			// holds true / false for each key

static int m_nFPS = 0;
static int m_nFrames = 0;											// FPS and FPS Counter
static bool g_pause=false;

cCamera g_cam;
clf_stage g_stage;
cCamera g_cam_instruments;
clf_stage g_stage_instuments;

// a light source
clf_light g_light(GL_LIGHT0);

// the shader
clf_shader shader;

// class to handle the fog effect
class clf_fog{
	GLfloat fogColor[4];
	GLfloat fogStart, fogEnd;
	GLfloat fogDensity;
public:
	clf_fog(){
		// set default parameters
		fogColor[0] = 0.55f;
		fogColor[1] = 0.55f;
		fogColor[2] = 0.55f;
		fogColor[3] = 1.f;
		fogStart = 0.1f;    // fog starts right from viewer's eyes
		fogEnd = 80.f;			// and the furthest you can see is 180 units
		fogDensity = .3f;	// default density
	}

	~clf_fog(){
	}

	void fog_apply(){
		glFogi(GL_FOG_MODE, GL_LINEAR);			// Fog Mode, almost natural, the further the object the deeper in the fog
		glFogfv(GL_FOG_COLOR, fogColor);		// Fog Color, here we can simulate artificial fog effect with color
		glFogf(GL_FOG_DENSITY, fogDensity);	// How Dense Will The Fog Be
		glFogf(GL_FOG_START, fogStart);			// Fog Start Depth
		glFogf(GL_FOG_END, fogEnd);					// Fog End Depth, nothing is visible beyond that limit
		glEnable(GL_FOG);										// Enable GL_FOG
	}

	void fog_terminate(){
		glDisable(GL_FOG);					// disable GL_FOG
	}

	GLfloat& start(){
		return fogStart;  // fog start distance
	}
	GLfloat& end(){
		return fogEnd;		// fog end distance
	}
	GLfloat& density(){
		return fogDensity;	// fog density
	}
	GLfloat& red(){
		return fogColor[0];		// red coordinate of fog color
	}
	GLfloat& green(){
		return fogColor[1];		// green coordinate of fog color
	}
	GLfloat& blue(){
		return fogColor[2];		// blue coordinate of fog color
	}
	GLfloat& alpha(){
		return fogColor[3];		// alpha coordinate of fog color
	}
};
clf_fog g_fog;

// base object type
class clf_world_object
{
public:
	static clf_vector3D gravity; // the famous g

	float m_mass;				// object mass
	float m_radius;     // bounding sphere
	clf_vector3D m_pos; // position
	clf_vector3D m_vel; // velocity

	clf_world_object(float m, float r ):m_mass(m), m_radius(r){
	}
	virtual ~clf_world_object(){
	}
	virtual void step_simulation(float fElapsed){
		m_pos = m_pos+m_vel * fElapsed  + fElapsed*fElapsed*0.5f*gravity;
		m_vel = m_vel+gravity*fElapsed;
	}
	virtual void render(){
	}
	// check for collision pure virtual function
	virtual bool collides_with(clf_world_object* pob)=0;
};
clf_vector3D clf_world_object::gravity = clf_vector3D(0.f, -9.81f, 0.f);

// our game world
class clf_world
{
	static float sim_speed; // used to control game speed, 1 is realtime <1 slow motion, >1 fast
public:
	// a list of the world objects
	std::list<clf_world_object*> m_objects;
	clf_world(){
	}
	~clf_world(){
	}
	// iterate the objects and find collisions
	void collision_detection(){
		std::list<clf_world_object*>::iterator it;
		for (it=m_objects.begin(); it!=m_objects.end(); ++it)
		{
			std::list<clf_world_object*>::iterator jt;
			jt = it;
			++jt;
			clf_world_object* iob = *it;
			for ( ; jt!=m_objects.end(); ++jt)
			{
				clf_world_object* job = *jt;
				iob->collides_with(job);
			}
		}
	}
	// calculate next frame
	void step_simulation(float fElapsed){
		std::list<clf_world_object*>::iterator it;
		fElapsed *= sim_speed;
		for (it=m_objects.begin(); it!=m_objects.end(); ++it)
		{
			// move objects
			(*it)->step_simulation(fElapsed);
		}
		// check for collisions
		collision_detection();
	}
	// draw the objects
	void render(){
		std::list<clf_world_object*>::iterator it;
		for (it=m_objects.begin(); it!=m_objects.end(); ++it)
		{
			(*it)->render();
		}
	}
	// add object to world
	void add_object(clf_world_object* pob){
		m_objects.push_back(pob);
	}
};
float clf_world::sim_speed = .2f;	// used to slow down (like slow motion) the simulation
// the game world
clf_world g_w;

// the sphere object
// enable/disable definition to see the simulation with gravity effect or not
#define NO_GRAVITY
class clf_sphere : public clf_world_object
{
public:
	clf_sphere(float m, float r ):clf_world_object(m,r){
	}
	virtual ~clf_sphere(){
	}
	virtual void render(){
		glPushMatrix();
		glTranslatef(m_pos.x, m_pos.y, m_pos.z);
		drawSphere(m_radius);
		glPopMatrix();
	}
	virtual void step_simulation(float fElapsed){
		// here we implement trajectory without gravity
#if defined NO_GRAVITY
		m_pos = m_pos+m_vel * fElapsed;
#else
		// base class implements gravity
		clf_world_object::step_simulation(fElapsed);
#endif
	}
	// calculate the result of the collision with another sphere
	void collision_respond(clf_sphere* pob){
		clf_vector3D x(pob->m_pos - m_pos);
		x.normalize();
		clf_vector3D u1x = x * dotProduct(x, m_vel);
		clf_vector3D u1y = m_vel - u1x;
		clf_vector3D u2x = -x * dotProduct(-x, pob->m_vel);
		clf_vector3D u2y = pob->m_vel - u2x;

		float e = .85f;
		float tot_mass = m_mass+pob->m_mass;
		clf_vector3D v1x = ((m_mass-e*pob->m_mass)/tot_mass)*u1x + ((1+e)*pob->m_mass/tot_mass)*u2x;
		clf_vector3D v2x = ((1+e)*m_mass/tot_mass)*u1x + ((pob->m_mass-e*m_mass)/tot_mass)*u2x;

		m_vel = u1y+v1x;
		pob->m_vel = u2y+v2x;

		x *= (m_radius + pob->m_radius);
		x *= 1.0001f;
		pob->m_pos = m_pos + x;
	}
	// check if the two spheres collide
	virtual bool collides_with(clf_sphere* pob){
		clf_vector3D dist_vec(pob->m_pos - m_pos);
		float dist = dist_vec.length();
		if (dist > (m_radius + pob->m_radius)	)
			return false;
		collision_respond(pob);
		return true;
	}
	// general collision detection function
	virtual bool collides_with(clf_world_object* pob){
		clf_sphere* ob = dynamic_cast<clf_sphere*>(pob);
		if (ob)
			return collides_with(ob);
		return true;
	}
};
// the two spheres of the game
clf_sphere g_s(1.8f, .8f);
clf_sphere g_s1(2.2f, 1.0f);

///////////////////////////////////////////////////////////////////////
void restart_game()
{
	// position the objects
	g_s.m_pos = clf_vector3D(-2.f, 1.5f, 0.f);
	g_s1.m_pos = clf_vector3D(2.f, 1.5f, -2.f);
	// set the speed 
	float speed = 8.f;
	// and launch
	g_s.m_vel = clf_vector3D(speed*cos(DegreesToRadians(45)),speed*sin(DegreesToRadians(45)),0);
	g_s1.m_vel = clf_vector3D(speed*cos(DegreesToRadians(135)),speed*sin(DegreesToRadians(135)),-speed*cos(DegreesToRadians(135)));
}

// this function is called when the program starts to initialize the game
void game_init()
{
	get_font_manager().create_font_2d("Bookman2D", "Bookman Old Style", 16);

	g_cam.setup(clf_vector3D(0,8,-35),       // position viewer
							clf_vector3D(2,2,0),        // setup the point of interest (where we are looking at)
							clf_vector3D(0,1,0));       // which way is up
	g_cam_instruments.setup(clf_vector3D(0,0,-1),       // position viewer
							clf_vector3D(0,0,0),        // setup the point of interest (where we are looking at)
							clf_vector3D(0,1,0));       // which way is up

	glEnable (GL_DEPTH_TEST);        // enable depth test
	glEnable(GL_COLOR_MATERIAL);     // use color for drawing the object

	// load shader
	shader.load_shaders("phong_vertex_shader2.glsl", "fragment_shader1.glsl");
	
	// Light model parameters:
	// -------------------------------------------
	GLfloat lm_ambient[] = {0.0, 0.0, 0.0, 0.0 };
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lm_ambient);
	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0);
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 0.0);

	// create and position a light source
	g_light.set_ambient(0.0f, 0.0f, 0.0f, 1.f);
	g_light.set_diffuse(0.7f, 0.7f, 0.7f, 1.f);
	g_light.set_specular(1.f, 1.f, 1.f, 1.f);
	g_light.set_position(-100.f, 100.f, -100.f, 0.f);

	// -------------------------------------------
	// Material parameters:
	
	GLfloat material_ambient[] = {0.7f, 0.7f, 0.7f, 1.0f};
	GLfloat material_diffuse[] = {0.4f, 0.4f, 0.4f, 1.0f};
	GLfloat material_specular[] = {0.8f, 0.8f, 0.8f, 1.0f};
	GLfloat material_emission[] = {0.1f, 0.1f, 0.1f, 0.0f};
	GLfloat material_shine = 50.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_ambient);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, material_emission);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material_shine);

	// add the objects to the world
	g_w.add_object(&g_s);
	g_w.add_object(&g_s1);

	restart_game();
}

// this function is called when the window is resized
void resize_window (int width, int height)			// reshape the window when it's moved or resized
{
	// setup viewports
	int hh = height/4;
	// main scene
	g_stage.resize(0,hh,width, height-hh);
	// instrument panel
	g_stage_instuments.resize(0,0,width, hh);
}

// this function is called upon exit to clear allocated resources
void game_terminate()
{
	shader.terminate();
}

// the core of the game. This is called repeatedly to update the state of the game
// fElapsed: is the elapsed time since last call. The time is in seconds (usualy a very small number)
static int source_mode=6;  // set initial values for transparency modes (HUD)
static int dest_mode=7;
static GLenum modes [] = {
	GL_ZERO,
	GL_ONE,
	GL_SRC_COLOR,
	GL_ONE_MINUS_SRC_COLOR,
	GL_DST_COLOR,
	GL_ONE_MINUS_DST_COLOR,
	GL_SRC_ALPHA,
	GL_ONE_MINUS_SRC_ALPHA,
	GL_DST_ALPHA,
	GL_ONE_MINUS_DST_ALPHA,
	GL_SRC_ALPHA_SATURATE,
};
int g_modes = sizeof(modes)/sizeof(GLenum);
static char* mode_names [] = {
	"GL_ZERO",
	"GL_ONE",
	"GL_SRC_COLOR",
	"GL_ONE_MINUS_SRC_COLOR",
	"GL_DST_COLOR",
	"GL_ONE_MINUS_DST_COLOR",
	"GL_SRC_ALPHA",
	"GL_ONE_MINUS_SRC_ALPHA",
	"GL_DST_ALPHA",
	"GL_ONE_MINUS_DST_ALPHA",
	"GL_SRC_ALPHA_SATURATE",
};
void frame_move(float fElapsed)
{
	if (keyDown[VK_ESCAPE]) 
		DestroyWindow(hWnd);
	if (keyDown['R'])
	{
		keyDown['R'] = 0;
		restart_game();
	}
	if (keyDown['P'])
	{
		keyDown['P'] = 0;
		g_pause = !g_pause;
	}
	if (keyDown['M'])
	{
		keyDown['M'] = 0;
		source_mode++;
		if (g_modes<=source_mode)
			source_mode = 0;
	}
	if (keyDown['N'])
	{
		keyDown['N'] = 0;
		dest_mode++;
		if (g_modes<=dest_mode)
			dest_mode = 0;
	}

	if (!g_pause)
	{
		g_w.step_simulation(fElapsed);
	}

	// frame counting mechanism
	// the higher the frame rate the faster the system
	// if frame rate drops below 30 we are in deep trouble
	// we should either optimize the program or buy a new computer
	static dFloat tot = 0;
	tot += fElapsed;
	m_nFrames++;												// Increment Our frame Counter
	if (tot >= 1.f)
	{
		tot = 0;
		m_nFPS = m_nFrames;
		m_nFrames = 0;

		char txt[200];
		sprintf(txt, "GusOnGames, fps:%d",m_nFPS);
		SetWindowText(hWnd, txt);
	}
}

// ground geometry
float calc_elev(float x, float y)
{
	float waveWidth = .05f;
	float waveHeight = 2.5f;
	float z = sin(x*waveWidth) * cos(y*waveWidth) * waveHeight;
 
	return z;
}

// the drawing function
// it is called right after the frame_move function
void frame_render()
{
	// clear the screen
	glClearColor (0.6f, 0.6f, 0.9f, 1.f);				// background colour
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// clear screen and depth buffer

	// main scene
#if 1
	glMatrixMode (GL_PROJECTION);									// select the projection matrix
	glLoadIdentity ();														// reset the projection matrix

	// viewport and perspective
	g_stage.prepare();

	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glPushMatrix();

	// position the viewer
	g_cam.apply();

	// position the light
	g_light.enable();
	g_light.apply();

	glEnable(GL_NORMALIZE);
	shader.begin();

	g_fog.fog_apply();

	static GLuint draw_list = -1;
	if (draw_list == -1)
	{
		draw_list = glGenLists(1);
		glNewList(draw_list, GL_COMPILE);
		glColor3f(0,0,1);
		float i,j;
		float step = 5.f;
		for (i=-10; i<11; i+=step)
		{
			for (j=-10; j<11; j+=step)
			{
				clf_vector3D e1(i, calc_elev(i,j),j), 
					e2(i,calc_elev(i,j+step),j+step), 
					e3(i+step,calc_elev(i+step,j+step),j+step), 
					e4(i+step,calc_elev(i+step,j),j), n1, n2;
				clf_getNormalVector(e1, e2, e3, n1);
				clf_getNormalVector(e1, e3, e4, n2);

				glBegin(GL_TRIANGLES);
				glNormal3f(n1.x,n1.y,n1.z);
				glVertex3f(e1.x, e1.y, e1.z);
				glVertex3f(e2.x, e2.y, e2.z);
				glVertex3f(e3.x, e3.y, e3.z);
				glEnd();
				
				glBegin(GL_TRIANGLES);
				glNormal3f(n2.x,n2.y,n2.z);
				glVertex3f(e1.x, e1.y, e1.z);
				glVertex3f(e3.x, e3.y, e3.z);
				glVertex3f(e4.x, e4.y, e4.z);
				glEnd();
			}
		}
		glEndList();
	}
	glCallList(draw_list);

	glColor3f(1,0,0);
	g_w.render();

	shader.end();
	glPopMatrix();
	glPopAttrib();	

	// draw the HUD screen
#if 1
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glPushMatrix();
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);

	glTranslatef(0,0,-1);
	glColor4f(.0f,1.f,1.f,.15f);
	glBlendFunc(modes[source_mode],modes[dest_mode]);	
	glEnable(GL_BLEND);
	glBegin(GL_QUADS);
	glNormal3f(0,0,1);
	glVertex3f(-.35f, -1.f, -1.f);
	glVertex3f(-.35f, -.15f, -1.f);
	glVertex3f(.35f, -.15f, -1.f);
	glVertex3f(.35f, -1.f, -1.f);
	glEnd();
	glDisable(GL_BLEND);

	glPopMatrix();
	glPopAttrib();
#endif

	// display HUD info
#if 1
	glMatrixMode (GL_PROJECTION);									// select the projection matrix
	glLoadIdentity ();														// reset the projection matrix

	// position the pilot in the cockpit
	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset
	glTranslatef(0,0,-1);

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);

	clf_font* gfnt = get_font_manager().get_font("Bookman2D");

	gfnt->use_screen_position(false);

	glColor4f(0.f,0.f,1.f,1.f);
	glRasterPos2f(-.3f, -.4f);
	gfnt->print_2d("source:%s",mode_names[source_mode]);

	glColor4f(1.f,0.f,1.f,1.f);
	glRasterPos2f(-.3f, -.5f);
	gfnt->print_2d("dest:%s",mode_names[dest_mode]);

	glPopAttrib();
#endif

#endif  // main scene


	// display imnstruments
#if 1
	// use scissor test to clear instrument viewport area
	glClearColor(0.1f, 0.1f, 0.1f, 1.0f );
	glScissor(g_stage_instuments.left, g_stage_instuments.bottom, g_stage_instuments.width, g_stage_instuments.height);
	glEnable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode (GL_PROJECTION);									// select the projection matrix
	glLoadIdentity ();														// reset the projection matrix
	// position the pilot in the cockpit
	g_stage_instuments.prepare();

	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	g_cam_instruments.apply();

	float x,y,z;
	x = 0;
	y = 0;
	z = 0;

	// use scissor to draw only within the stick position instrument
	y = (float)g_stage_instuments.height;
	x = (g_stage_instuments.width-y)/2;
	glClearColor(0.1f, 0.1f, 0.6f, 1.0f );
	glScissor((GLsizei)x, (GLsizei)0, (GLsizei)y, (GLsizei)y);
	// clear the stick position display area
	glClear(GL_COLOR_BUFFER_BIT);

	// draw the cross
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_LINES);
		glVertex3f(-.35f, 0.f,0.f);
		glVertex3f(.35f, 0.f,0.f);
		glVertex3f(0.f, -.35f,0.f);
		glVertex3f(0.f, .35f,0.f);
	glEnd();
	// and the stick position indicator
	glColor3f(.75f, .15f, .15f);
	glBegin(GL_QUADS);
		glVertex3f(0.04f, 0.f,0.f);
		glVertex3f(0.f, 0.04f,0.f);
		glVertex3f(-0.04f, 0.f,0.f);
		glVertex3f(0.f, -0.04f,0.f);
	glEnd();
	glDisable(GL_SCISSOR_TEST);
#endif
}

